structure Ast =
struct

datatype prim_op = 
 OP_PLUS | OP_MINUS | OP_MUL | OP_DIV |
 OP_EQUAL | OP_NEQ | OP_LEQ | OP_GEQ |
 OP_GREATER | OP_LESS |
 (* TODO *)
 OP_MOD | OP_AND | OP_OR | OP_NOT |
 OP_TILDE
 
datatype pre_post_op =
 OP_INCR | OP_DECR

datatype ty =
    TVoid
  | TInt
  | TUInt
  | TChar
  | TPtr of ty
  | TStruct of string
  | TArray of ty * int option
  | TStructField of ty (* auxilary, similar to C++ ptr-to-member *)
  
type term = string

datatype var_kind = Param | Local | Global

type vctx = (string * (var_kind * ty)) list

type struct_info = {
 name : string,
 fields : (string * ty) list
}

datatype exp =
    EVar of string
  | ECchar of char
  | ECint of int
  | ECuint of int
  | ENull
  | EDeref of exp
  | EAddr of exp
  | EDot of exp * string
  | EArrow of exp * string
  | EIndex of exp * exp
  | EAssign of exp * exp
  | EPre of pre_post_op * exp
  | EPost of pre_post_op * exp
  | EPrim1 of prim_op * exp
  | EPrim2 of prim_op * exp * exp
  | EFree of exp
  | EFreeArray of exp * exp
  | ENew of ty
  | ENewArray of ty * exp
  | EAnd of exp * exp
  | EOr of exp * exp
  | ECall of string * exp list

datatype ghost =
  GAssign of string * string

datatype sm =
    SSkip
  | SSeq of sm * sm
  | SExec of exp
  | SBlock of (string * ty) list * sm
  | SIf of exp * sm * sm
  | SWhile of string option * exp * sm
  | SBreak
  | SContinue
  | SReturn of exp
  | SGhost of ghost

datatype any =
    SM of sm
  | EXP of exp
  | DECLS of (string * ty) list
  | STRUCT_DEF of struct_info
  
end