theory MapExt
imports Main
begin

subsection {* Auxiliary Lemmata about Sets *}

lemmas set_eq_subsetD1 =
  set_eq_subset[THEN iffD1, THEN conjunct1, THEN subsetD]

lemmas set_eq_subsetD2 =
  set_eq_subset[THEN iffD1, THEN conjunct2, THEN subsetD]

lemma set_disjoint_ccontr:
  "\<lbrakk> A \<inter> B = {}; x \<in> A; x \<in> B \<rbrakk> \<Longrightarrow> False"
  by blast

lemma set_disjoint_ccontr_dom:
  "\<lbrakk> dom A \<inter> dom B = {}; A x = Some y; B x = Some y' \<rbrakk> \<Longrightarrow> False"
  by blast

lemma set_disjoint_subset:
  "\<lbrakk> A \<inter> B = {}; A' \<subseteq> A; B' \<subseteq> B \<rbrakk> \<Longrightarrow> A' \<inter> B' = {}"
  by blast

lemma disjoint_leftD: 
  "\<lbrakk> A \<inter> B = {}; x \<in> A \<rbrakk> \<Longrightarrow> x \<notin> B"
  by blast
  
lemma disjoint_setsI:
  "\<lbrakk> \<And>x. \<lbrakk> x \<in> A; x \<in> B \<rbrakk> \<Longrightarrow> False \<rbrakk> \<Longrightarrow> A \<inter> B = {}"
  by auto

subsection {* Additional Lemmata about Maps *}

notation
  map_le (infix "\<le>m" 50)

lemma map_leD:
  "\<lbrakk> f \<le>m g;
     f x = Some y
   \<rbrakk> \<Longrightarrow> g x = Some y"
  apply (clarsimp simp add: map_le_def)
  apply (drule bspec, rule domI, assumption)
  apply simp
  done

lemma map_le_restrict[simp,intro]:
  "f |` X \<le>m f"
  by (auto simp add: map_le_def restrict_map_def)

lemma map_le_restrictD:
  "f \<le>m g \<Longrightarrow> f = g |` dom f"
  by (auto intro!: ext simp add: map_le_def restrict_map_def)

lemma map_le_finite_domD:
  "\<lbrakk> finite (dom g); f \<le>m g \<rbrakk> \<Longrightarrow> finite (dom f)"
  apply (rule finite_subset[rotated], assumption)
  apply (rule map_le_implies_dom_le, assumption)
  done

lemma map_restrict_add:
  "f |` A ++ f |` B = f |` (A \<union> B)"
  by (auto intro!: ext simp add: restrict_map_def map_add_def split: option.split)

lemma map_add_restrict:
  "(f ++ g) |` A = f |` A ++ g |` A"
  by (auto intro!: ext simp add: restrict_map_def map_add_def split: option.split)

lemma map_restrict_sub_dom:
  "dom f \<subseteq> A \<Longrightarrow> f |` A = f"
  apply (unfold dom_def restrict_map_def)
  apply (auto intro!: ext)
  apply (drule(1) contra_subsetD)
  apply auto
  done

lemma map_restrict_dom[simp]:
  "f |` dom f = f"
  by (simp add: map_restrict_sub_dom)

lemma map_upd_map_add_conv:
  "A(x \<mapsto> y) = A ++ [x \<mapsto> y]"
  by simp

lemma case_by_outcome[simp]:
  "((if P then x else None) = Some y) = (P \<and> x = Some y)"
  "((if P then x else None) = None) = (\<not> P \<or> x = None)"
  "((if P then None else x) = Some y) = (\<not> P \<and> x = Some y)"
  "((if P then None else x) = None) = (P \<or> x = None)"
  by auto

lemma not_dom_simps[simp]:
  "x \<notin> dom f \<Longrightarrow> (f x = Some y) = False"
  "x \<notin> dom f \<Longrightarrow> (f x = None) = True"
  "x \<notin> dom g \<Longrightarrow> (f ++ g) x = f x"
  by (auto simp add: dom_def map_add_def)

lemma map_restrict_Some[simp]:
  "((G |` D) q = Some r) = (G q = Some r \<and> q \<in> D)"
  by (auto simp add: restrict_map_def)

lemma restrict_map_singleton[simp]:
  "f x = Some y \<Longrightarrow> f |` {x} = [x \<mapsto> y]"
  by (auto intro: ext simp add: restrict_map_def)

lemma map_restrict_UNIV[simp]:
  "f |` UNIV = f"
  by (simp add: restrict_map_def)

lemma map_add_restrict_dom[simp]:
  "(A ++ C) |` dom C = C"
  apply (rule ext)
  apply (auto simp add: restrict_map_def map_add_def split: option.split)
  done

lemma map_restrict_Diff[simp]:
  "f |` (A - B) = f |` A |` (- B)"
  by (auto intro!: ext simp add: restrict_map_def)

lemma map_restrict_add_compl[simp]:
  "f |` A ++ f |` (- A) = f"
  by (auto intro!: ext simp add: restrict_map_def map_add_def split: option.split)

lemma map_add_restrict'[simp]:
  "(A ++ B) |` (- dom A) = B |` (- dom A)"
  apply (unfold restrict_map_def map_add_def)
  apply (auto intro!: ext split: option.split_asm)
  done

lemma map_restrict_setminus_absorb:
  "dom f \<inter> A = {} \<Longrightarrow> f |` (- A) = f"
  apply (unfold restrict_map_def)
  apply (auto intro!: ext split: option.split_asm)
  apply (rule ccontr, drule not_sym)
  apply fast
  done

lemma dom_eqD:
  "\<lbrakk> f x = Some y; dom f = D \<rbrakk> \<Longrightarrow> x \<in> D"
  by (unfold dom_def, auto)

definition
  "map_graph f = { (x,y). f x = Some y }"

lemma map_graph_empty[simp]:
  "map_graph empty = {}"
  by (simp add: map_graph_def)

lemma map_graph_single[simp]:
  "map_graph [ x \<mapsto> y ] = {(x,y)}"
  by (auto simp add: map_graph_def split: split_if_asm)

lemma map_graph_add:
  "dom A \<inter> dom B = {} \<Longrightarrow> map_graph (A ++ B) = map_graph A \<union> map_graph B"
  apply (auto simp add: map_graph_def)
  apply (subst not_dom_simps(3)) 
  apply auto
  done

lemma map_graph_insert:
  "A x = Some y \<Longrightarrow> map_graph A = insert (x, y) (map_graph (A |` (- {x})))"
  by (auto simp add: map_graph_def)

lemma map_graph_insert':
  "map_graph (A(x\<mapsto>y)) = insert (x, y) (map_graph (A |` (- {x})))"
  by (auto simp add: map_graph_def map_graph_insert split: split_if_asm)

lemma map_graph_finite[simp]:
  "finite (map_graph f) = finite (dom f)"
  apply (auto simp add: map_graph_def)
  apply (unfold dom_def)
  apply (drule_tac f=fst in finite_image_set)
  apply simp
  apply simp
  apply (drule_tac f="\<lambda>x. (x, the (f x))" in finite_image_set)
  apply (erule arg_cong[where f=finite, THEN iffD2, rotated])
  apply auto
  done

lemma map_graph_fst_image[simp]:
  "fst ` map_graph A = dom A"
  apply (unfold map_graph_def image_def dom_def)
  apply auto
  done

subsection {* Quantification over Map Entries *}

definition
  AllMap :: "('a \<rightharpoonup> 'b) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> bool"
where
  "AllMap f P \<equiv> \<forall>x y. f x = Some y \<longrightarrow> P x y"

syntax
  "_AllMap" :: "'a \<Rightarrow> 'b \<Rightarrow> ('a \<rightharpoonup> 'b) \<Rightarrow> ('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> bool" ("\<forall> _ \<mapsto> _ \<in> _. _" [0,0,0,10] 10)

translations
  "\<forall>x \<mapsto> y \<in> f. P" == "CONST AllMap f (\<lambda>x y. P)"

lemma all_mapD[dest]:
  "\<lbrakk> \<forall>x \<mapsto> y \<in> f. P x y; f x = Some y \<rbrakk> \<Longrightarrow> P x y"
  by (simp add: AllMap_def)

text {* Parallels bspec in Set.thy *}

declaration {*
 K (Classical.map_cs (fn cs => cs addbefore ("all_map_spec", dtac @{thm all_mapD} THEN' assume_tac)))
*}

lemma all_mapI[intro!]:
  "\<lbrakk> \<And>x y. f x = Some y \<Longrightarrow> P x y
   \<rbrakk> \<Longrightarrow> \<forall>x \<mapsto> y \<in> f. P x y"
  by (simp add: AllMap_def)

lemma all_map_cong[cong]:
  "\<lbrakk> f = f'; \<And>x y. f' x = Some y \<Longrightarrow> P x y = P' x y
  \<rbrakk> \<Longrightarrow> (AllMap f P) = (AllMap f' P')"
  by (simp add: AllMap_def)

lemma all_map_empty[simp,intro!]:
  "\<forall>x \<mapsto> y \<in> Map.empty. P x y"
  by (simp add: AllMap_def)

lemma all_map_upd:
  "(\<forall>x\<mapsto>y \<in> f(x' \<mapsto> y'). P x y) = (P x' y' \<and> (\<forall>x\<mapsto>y \<in> f |` (-{x'}). P x y))"
  by (auto simp add: AllMap_def)
  
lemma all_map_add:
  "(\<forall>x\<mapsto>y \<in> (f ++ g). P x y) = ((\<forall>x\<mapsto>y\<in>g. P x y) \<and> (\<forall>x\<mapsto>y\<in>(f |` (- dom g)). P x y))"
  by (fastforce simp add: AllMap_def map_add_def split: option.split_asm)

lemma all_map_single[simp]:
  "(\<forall>x\<mapsto>y\<in>[x \<mapsto> y]. P x y) = P x y"
  by (simp add: AllMap_def)

text {* Enable the "named values" approach 
  for reasoning about abstract values *}
lemma all_map_upd'[simp]:
  "x \<notin> dom f \<Longrightarrow> (\<forall>x\<mapsto>y \<in> (f(x \<mapsto> y)). P x y) = (P x y \<and> (\<forall>x\<mapsto>y\<in>f. P x y))"
  by (simp add: all_map_upd map_restrict_setminus_absorb)

lemma all_map_add'[simp]:
  "dom f \<inter> dom g = {} 
   \<Longrightarrow> (\<forall>x\<mapsto>y\<in>(f ++ g). P x y) = ((\<forall>x\<mapsto>y\<in> g. P x y) \<and> (\<forall>x\<mapsto>y\<in>f. P x y))"
  by (simp add: all_map_add map_restrict_setminus_absorb Int_ac)

lemma map_all_True[simp,intro]:
  "(\<forall> x \<mapsto> y \<in> N. True) = True"
  by (simp add: AllMap_def)

lemma all_map_dom_eqE:
  "\<lbrakk> \<forall> x \<mapsto> y \<in> f. P x y; dom f = A;
     \<And>x y. \<lbrakk> x \<in> A \<longrightarrow> f x = Some y \<longrightarrow> P x y \<rbrakk> \<Longrightarrow> R
   \<rbrakk> \<Longrightarrow> R"
  by (auto simp add: AllMap_def)

lemma all_map_singleton:
  "x' \<in> dom f \<Longrightarrow> (\<forall> x \<mapsto> y \<in> f. x = x' \<longrightarrow> P x y) = (\<exists>y. f x' = Some y \<and> P x' y)"
  "x' \<in> dom f \<Longrightarrow> (\<forall> x \<mapsto> y \<in> f. x' = x \<longrightarrow> P x y) = (\<exists>y. f x' = Some y \<and> P x' y)"
  by (auto simp add: AllMap_def)

lemmas all_map_singletonD =
  all_map_singleton[THEN iffD1, rotated 1, standard]

lemma map_add_select[simp]:
  "x \<in> dom f' \<Longrightarrow> (f ++ f') x = f' x"
  "x \<notin> dom f' \<Longrightarrow> (f ++ f') x = f x"
  apply (simp_all add: map_add_def)
  apply (auto split: option.split_asm)
  done

(* parallels domIff, an equivalent of which is in the clarset *)
lemma ranIff[iff,simp del]:
  "(y \<in> ran f) = (\<exists>x. f x = Some y)"
  by (auto simp add: ran_def)

lemma dom_by_ran_elemD:
  "\<lbrakk> ran f = ys; y \<in> ys \<rbrakk> \<Longrightarrow> \<exists>x. f x = Some y"
  by auto
  
lemma f_image_dom_f[simp]:
  "(f ` dom f) = Some ` ran f"
  by (fastforce intro: sym simp add: image_def)
  
lemma map_skip_right[simp]:
  "n x = None \<Longrightarrow> (m ++ n) x = m x"
  by (simp add: map_add_def)

lemma ran_map_add_disjoint:
  "dom f \<inter> dom g = {} \<Longrightarrow> ran (f ++ g) = ran f \<union> ran g"
  apply (auto simp add: ran_def map_add_def split: option.split)
  apply (rule_tac x=a in exI)
  apply simp
  done

subsection {* Injective maps *}

definition
  "inj_map f \<equiv> inj_on f (dom f)"

lemma inj_map_empty[simp,intro]:
  "inj_map empty"
  by (simp add: inj_map_def)

text {*
 The original node of some pointer is expressed directly using the inv combinator.
*}
definition
  "inv_map f p = Some (the_inv_into (dom f) f (Some p))"

lemma inv_map_f_y: 
  "\<lbrakk>inj_map f; f x = Some y \<rbrakk> \<Longrightarrow> inv_map f y = Some x"
  by (fastforce dest:  the_inv_into_f_f[OF _ domI] simp add: inv_map_def inj_map_def)

lemma inv_map_f_f:
  "\<lbrakk> inj_map f; x \<in> dom f \<rbrakk> \<Longrightarrow> inv_map f (the (f x)) = Some x"
  by (auto simp add: inv_map_f_y)

lemma Some_image[simp]:
  "(Some x \<in> Some ` A) = (x \<in> A)"
  by auto
  
lemma inj_map_singleton[simp,intro!]:
  "inj_map [x \<mapsto> y]"
  by (simp add: inj_map_def)

lemma f_inv_map_f:
  "\<lbrakk> inj_map f; y \<in> ran f \<rbrakk> \<Longrightarrow> f (the (inv_map f y)) = Some y"
  by (fastforce intro: f_the_inv_into_f simp add: inv_map_def inj_map_def)


(* analogous: inj_on_map_add_dom *)
lemma inj_map_add_disjoint:
  "\<lbrakk> dom f \<inter> dom g = {}; ran f \<inter> ran g = {} \<rbrakk> \<Longrightarrow> inj_map (f ++ g) = (inj_map f \<and> inj_map g)"
  apply (auto simp add: inj_map_def)
  apply (clarsimp intro!: inj_onI)
  apply (drule_tac x=x and y=y in inj_onD)
  apply (subgoal_tac "x \<notin> dom g \<and> y \<notin> dom g")
  apply (simp add: map_add_select)
  apply (blast dest: set_disjoint_ccontr_dom)
  apply blast
  apply blast
  apply blast
  apply (clarsimp intro!: inj_onI)
  apply (drule_tac x=x and y=y in inj_onD)
  apply fastforce+

  apply (clarsimp intro!: inj_onI)
  apply (erule HOL.disjE)
  apply (subgoal_tac "y \<in> dom g")
  apply simp
  apply (drule(3) inj_onD, simp)
  apply (rule ccontr)
  apply clarsimp
  apply (rename_tac z)
  apply (subgoal_tac "z \<in> ran f \<and> z \<in> ran g")
  apply fast
  apply fast
  
  apply (subgoal_tac "x \<notin> dom g \<and> y \<notin> dom g \<and> y \<in> dom f")
  apply clarsimp
  apply (drule_tac x=x and y=y in inj_onD, simp, fast, fast) 
  apply fast
  apply (simp add: disjoint_leftD)
  apply (rule context_conjI)
  prefer 2
  apply (simp add: disjoint_leftD)
  apply (rule ccontr)
  apply clarsimp
  apply (rename_tac z)
  apply (subgoal_tac "z \<in> ran f \<and> z \<in> ran g")
  apply fast
  apply fast
  done

(* Formulate as splitting rule useful in separation logic? *)
lemma inj_map_upd_disjoint[simp]:
  "\<lbrakk> x \<notin> dom f; y \<notin> ran f \<rbrakk> \<Longrightarrow> inj_map (f(x \<mapsto> y)) = inj_map f"
  apply (rule trans)
  apply (rule inj_map_add_disjoint[where f=f and g="[x \<mapsto> y]", simplified])
  apply auto
  done

lemma inj_map_map_le:
  "\<lbrakk> inj_map g; f \<le>m g \<rbrakk> \<Longrightarrow> inj_map f"
  apply (simp add: inj_map_def)
  apply (rule inj_onI)
  apply clarsimp
  apply (drule_tac x=x and y=y in inj_onD)
  apply (auto dest: map_leD)
  apply (metis map_leD)
  done

lemma inj_mapD:
  "\<lbrakk> inj_map f; f x = Some y; f x' = Some y \<rbrakk> \<Longrightarrow> x = x'"
  apply (simp add: inj_map_def)
  apply (drule_tac x=x and y=x' in inj_onD)
  apply (auto intro: domI)
  done

lemma inv_map_map_le:
  "\<lbrakk> \<phi> \<le>m \<phi>'; inj_map \<phi>'; p \<in> ran \<phi> \<rbrakk> \<Longrightarrow> inv_map \<phi> p = inv_map \<phi>' p"
  apply (frule(1) inj_map_map_le)
  apply (clarsimp simp add: inv_map_def the_inv_into_def ranIff domI cong: rev_conj_cong)
  apply (rule_tac a=x in theI2)
  apply (fastforce dest: map_le_implies_dom_le map_leD intro: domI)
  apply (fastforce dest: inj_mapD map_leD)
  apply (rule_tac a=x in theI2)
  apply (fastforce dest: map_le_implies_dom_le map_leD intro: domI)
  apply (fastforce dest: inj_mapD map_leD)
  apply (fastforce dest: inj_mapD map_leD)
  done

lemma dom_of_cases[simp]:
  "dom (\<lambda>p. if P p then Some (a p) else None) = {x. P x}"
  by auto

lemma ran_of_cases[simp]:
  "ran (\<lambda>p. if P p then Some (a p) else None) = (\<lambda>p. a p) ` {x. P x}"
  by auto
  
(* TODO: propose for the library *)
lemma ranD:
  "\<lbrakk> ran f = A; x \<in> dom f \<rbrakk> \<Longrightarrow> the (f x) \<in> A"
  by auto 

lemma ranD':
  "\<lbrakk> ran f = A; f x = Some y \<rbrakk> \<Longrightarrow> y \<in> A"
  by (auto simp add: ranIff)

lemma the_in_ran:
  "x \<in> dom f \<Longrightarrow> the (f x) \<in> ran f"
  by auto

lemma the_Some[simp]:
  "the \<circ> Some = (\<lambda>x. x)"
  by (simp add: comp_def)

end