structure Decls =
struct

open Ast

datatype type_specifier =
    TypeName of string 
  | Struct of string

datatype direct =
    Name of string
  | Array of direct * exp option
  | Nested of declarator
and declarator =
    Declarator of int * direct

type declaration_specifiers = type_specifier list

fun tr_specifiers specs =
  case specs of
      [ TypeName("int") ] => TInt
    | [ TypeName("void") ] => TVoid
    | [ Struct name ] => TStruct name
    | s => error ("Unknown named type specifiers "^PolyML.makestring specs)

fun add_pointers 0 ty = ty
  | add_pointers i ty = add_pointers (i-1) (TPtr ty)
  
fun tr_type specs ptrs =
  add_pointers ptrs (tr_specifiers specs)

fun eval (ECint i) = i
  | eval exp = error ("Cannot evaluate constant "^PolyML.makestring exp)

fun tr_declarator specs (Declarator(ptrs,direct)) =
  let fun loop (Name x) = (x, tr_type specs ptrs)
        | loop (Array(decl,n)) =
          (let val (x,ty) = loop decl
               val n' = case n of
                            NONE => NONE
                          | SOME nn => SOME (eval nn)
           in
               (x, add_pointers ptrs (TArray(ty, n')))
           end)
        | loop (Nested decl) =
          (let val (x,ty) = tr_declarator specs decl
           in
               (x, add_pointers ptrs ty)
           end)
  in
      loop direct
  end

fun make_decl (specs : declaration_specifiers) (decl : declarator) =
  tr_declarator specs decl

end