structure InstDefs =
struct

fun instantiate_defs_tac ctx =
    SUBGOAL (fn (t,i) =>
    let val str = Syntax.string_of_term ctx
        fun cert t = cterm_of (Proof_Context.theory_of ctx) t
          handle (CTERM _) => error ("certify: "^str t)
        exception DEF_FAIL
        fun mk' bs x args y =
          let val str_t = Syntax.string_of_term ctx
              fun trans d (t as Const _) = t
                | trans d (t as Free _) = t
                | trans d (t as Var _) = t
                | trans d (Bound i) =
                  if i < d
                  then t
                  else
                (case find_index (fn b => b = Bound (i - d)) (rev args) of
                     ~1 => (warning ("Binding "^str_t (Var x)^"->"^str_t y^
                                     "\nMissing bound var "^string_of_int (i - d)^
                                     " = "^fst (nth bs (i - d))^" in ("^
                                     commas (map str args)^")");
                            raise DEF_FAIL)
                   | j => Bound (d+j))
                | trans d (Abs(x,T,b)) = Abs(x,T, trans (d+1) b)
                | trans d (a $ b) = trans d a $ trans d b

              val params = map_index (fn (i,a) =>
                                        let val T = case a of
                                                        Bound j => snd (nth bs j)
                                                      | _ => fastype_of a
                                        in
                                            ("x"^string_of_int i, T)
                                        end) args
              val r = Logic.rlist_abs (params, trans 0 y)
              (* val _ = tracing ("R "^str (Var x)^"->"^str r) *)
          in
              (cert (Var x), cert r)
          end

        fun mk bs x y res =
          ((*tracing ("mk "^str x^" = "^str y); *)
          case strip_comb x of
              (Var x, args) => mk' bs x args y :: res
            | _ => case strip_comb y of
                (Var y, args) => mk' bs y args x :: res
                     | _ => res)
          handle DEF_FAIL => res

        fun insts bs t res =
          case t of
              (Const(@{const_name "HOL.eq"}, _) $ x $ y) => mk bs x y res
            | (Abs(x,T,b)) => insts ((x,T) :: bs) b res
            | (a $ b) => insts bs a (insts bs b res)
            | _ => res

        val refl_ss =
           Simplifier.clear_ss (simpset_of ctx) addsimps
           (@{thms simp_thms} @ [ @{thm refl} ])
    in
      PRIMITIVE (Thm.instantiate ([], insts [] t [])) THEN
      simp_tac (simpset_of ctx) i
    end)

val setup =
    Method.setup
      (Binding.name "inst_defs")
      (Scan.succeed (fn ctx => (Method.SIMPLE_METHOD (instantiate_defs_tac ctx 1))))
      "instatiate all equalities directly"

end