structure SimplC =
struct

open Unsynchronized

structure SimplCLrVals =
SimplCLrValsFun(structure Token = LrParser.Token)

structure SimplCLex =
SimplCLexFun(structure Tokens = SimplCLrVals.Tokens);

structure SimplCParser =
Join(structure LrParser = LrParser
     structure ParserData = SimplCLrVals.ParserData
     structure Lex = SimplCLex)

fun invoke lexstream =
  let fun print_error (s,pos:Pos.t,_) =
          raise Pos.ParseError (pos,s)
  in 
      SimplCParser.parse(0,lexstream,print_error,())
  end;

fun makeStringLexer str =
    let val read = ref false
    in
        SimplCParser.makeLexer
          (fn _ => 
              if !read
              then ""
              else (read := true; str))
    end

fun parseWithLexer startSymbol lexer =
  let val dpos = Pos.dummy
      val lexer =
       SimplCParser.Stream.cons (startSymbol (dpos,dpos), lexer)
      val (result,lexer) = invoke lexer
      val (nextToken,lexer) = SimplCParser.Stream.get lexer
      val dummyEOF = SimplCLrVals.Tokens.EOF(Pos.dummy,Pos.dummy)
      val (SimplCParser.Token.TOKEN(_,(_,nextPos,_))) = nextToken
  in
      if not (SimplCParser.sameToken(nextToken,dummyEOF))
      then raise (Pos.ParseError(nextPos,"could not proceed"))
      else result
  end

fun dumpResult result = 
  (TextIO.print "OK ";
   TextIO.print (PolyML.makestring result);
   ())
   

val startSymbolSM = SimplCLrVals.Tokens.S_SM
val startSymbolSM_SEQ = SimplCLrVals.Tokens.S_SM_SEQ
val startSymbolEXP = SimplCLrVals.Tokens.S_EXP
val startSymbolDECLS = SimplCLrVals.Tokens.S_DECLS
val startSymbolSTRUCT = SimplCLrVals.Tokens.S_STRUCT

fun parseString startSymbol =
   makeStringLexer #>
   parseWithLexer startSymbol
   
fun parseAndDumpStringSM s =
   (s |>
    parseString startSymbolSM |>
    dumpResult)
   handle(Pos.ParseError (pos,msg)) =>
      (TextIO.print ("ERROR "^Pos.fmt pos^":"^msg^"\n");
       OS.Process.exit OS.Process.failure)


fun parseAndDumpStin () =
    ((SimplCParser.makeLexer 
                     (fn _ =>
                          case TextIO.inputLine TextIO.stdIn of
                              NONE => ""
                            | SOME line => line)) |>
    parseWithLexer startSymbolSM |>
    dumpResult)
   handle(Pos.ParseError (pos,msg)) =>
      (TextIO.print ("ERROR "^Pos.fmt pos^":"^msg^"\n");
       OS.Process.exit OS.Process.failure)

fun parseExpression str =
  let val (Ast.EXP decls) = parseString startSymbolEXP str
  in 
      decls
  end
  handle (Pos.ParseError(pos, msg)) =>
     error (Pos.fmt pos^":"^msg)

fun parseStatement str =
  let val (Ast.SM decls) = parseString startSymbolSM_SEQ str
  in 
      decls
  end
  handle (Pos.ParseError(pos, msg)) =>
     error (Pos.fmt pos^":"^msg)

fun parseDecls str =
  let val (Ast.DECLS decls) = parseString startSymbolDECLS str
  in 
      decls
  end
  handle (Pos.ParseError(pos, msg)) =>
     error (Pos.fmt pos^":"^msg)

fun parseStructDef str =
  let val (Ast.STRUCT_DEF def) = parseString startSymbolSTRUCT str
  in 
      def
  end
  handle (Pos.ParseError(pos, msg)) =>
     error (Pos.fmt pos^":"^msg)

end (* struct Simplc *)