structure Tokens = Tokens

type pos = Pos.t
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult= (svalue,pos) token

val dpos = Pos.dummy

val file = ref Pos.dummy_file
val line = ref 0
val line_start = ref 0

fun newline yypos =
  (line := (!line) + 1;
   line_start := yypos)

fun pos yypos =
  Pos.mk (!file) (!line) (yypos - !line_start)

val eof = fn () => Tokens.EOF(dpos,dpos)
fun error yypos msg = raise (Pos.ParseError(pos yypos,msg));

fun wrap tok yypos = tok (pos yypos,dpos)

val assn_text = ref [];

fun keep_assn yytext =
  assn_text := String.sub (yytext,0) :: !assn_text;

val keyword_tab =
  Symtab.make 
  [ ("while", Tokens.WHILE),
    ("if", Tokens.IF),
    ("else", Tokens.ELSE),
    ("do", Tokens.DO),
    ("for", Tokens.FOR),
    ("break", Tokens.BREAK),
    ("continue", Tokens.CONTINUE),
    ("struct", Tokens.STRUCT),
    ("enum", Tokens.ENUM),
    ("null", Tokens.NULL) ];

val type_id_tab =
  Symtab.make 
  [ ("void",()),
    ("int", ()),
    ("short", ()),
    ("long", ()),
    ("float", ()),
    ("double", ()),
    ("bool", ()),
    ("signed", ()),
    ("unsigned", ()) ];

%%
%header (functor SimplCLexFun(structure Tokens: SimplC_TOKENS));
%s ASSN GHOST;
alpha=[A-Za-z];
digit=[0-9];
ws = [\ \t];
idchar=[A-Za-z0-9_];
%%
<INITIAL>\n => (newline yypos; lex());
<INITIAL>{ws}+ => (lex());
<INITIAL>{digit}+ => (Tokens.LINT(the (Int.fromString yytext),pos yypos,dpos));
<INITIAL>"/*@" => (YYBEGIN ASSN; assn_text := []; lex());
<INITIAL>"//@" => (YYBEGIN GHOST; assn_text := []; lex());
<INITIAL>"++" => (wrap Tokens.PPLUS yypos);
<INITIAL>"--" => (wrap Tokens.MMINUS yypos);
<INITIAL>"+" => (wrap Tokens.PLUS yypos);
<INITIAL>"-" => (wrap Tokens.MINUS yypos);
<INITIAL>"*" => (wrap Tokens.MUL yypos);
<INITIAL>"/" => (wrap Tokens.DIV yypos);
<INITIAL>"=" => (wrap Tokens.ASSIGN yypos);
<INITIAL>"==" => (wrap Tokens.EQUAL yypos);
<INITIAL>"!=" => (wrap Tokens.NEQ yypos);
<INITIAL>"<" => (wrap Tokens.LESS yypos);
<INITIAL>">" => (wrap Tokens.GREATER yypos);
<INITIAL>"<=" => (wrap Tokens.LEQ yypos);
<INITIAL>">=" => (wrap Tokens.GEQ yypos);
<INITIAL>"->" => (wrap Tokens.ARROW yypos);
<INITIAL>"," => (wrap Tokens.COMMA yypos);
<INITIAL>";" => (wrap Tokens.SEMI yypos);
<INITIAL>"(" => (wrap Tokens.LPAR yypos);
<INITIAL>")" => (wrap Tokens.RPAR yypos);
<INITIAL>"[" => (wrap Tokens.LBRAC yypos);
<INITIAL>"]" => (wrap Tokens.RBRAC yypos);
<INITIAL>"{" => (wrap Tokens.LCURL yypos);
<INITIAL>"}" => (wrap Tokens.RCURL yypos);
<INITIAL>"&&" => (wrap Tokens.DAND yypos);
<INITIAL>"||" => (wrap Tokens.DOR yypos);
<INITIAL>{alpha}{idchar}* => 
(case Symtab.lookup keyword_tab yytext of
   SOME tok => wrap tok yypos
 | NONE =>
 (if Symtab.defined type_id_tab yytext
  then Tokens.TYID(yytext,pos yypos,dpos)
   else Tokens.ID(yytext,pos yypos,dpos)));
<INITIAL>. => (error yypos ("unexpected character "^yytext)); 

<ASSN>"*/" => (YYBEGIN INITIAL; 
               let val assn = String.implode (rev (!assn_text))
               in
                  Tokens.ASSN(assn,pos yypos,dpos)
               end);
<ASSN>\n => (newline yypos; keep_assn yytext; lex());
<ASSN>. => (keep_assn yytext; lex());

<GHOST>\n => (newline yypos;
              YYBEGIN INITIAL; 
              let val ghost = String.implode (rev (!assn_text))
              in
                Tokens.GHOST(ghost,pos yypos,dpos)
              end);
<GHOST>. => (keep_assn yytext; lex());
