structure SimplC_Goals =
struct

(**
  Switch between heap models
*)
val (heap_model, heap_model_setup) =
  Attrib.config_string @{binding "heap_model"} (K "")

datatype heap_models = Burstall | SepLogic

fun get_heap_model ctx =
  case Config.get ctx heap_model of
    "burstall" => Burstall
   | "seplogic" => SepLogic
   | "" => error "Heap model not set: choose 'burstall' or 'seplogic'"
   | m => error ("Unknown heap model "^m^": choose 'burstall' or 'seplogic'")

fun is_burstall ctx =
  get_heap_model ctx = Burstall

fun is_sep_logic ctx =
  get_heap_model ctx = SepLogic

val is_sep_logic_global = is_sep_logic o Proof_Context.init_global

(**
 Parameters -- towards a later generalization as a functor
*)
fun param_HOL_number_of_int i =
  HOLogic.mk_number @{typ int} i

fun cnst nm = Const(nm,dummyT)

val raw_tab = [
  ("prim_plus_int",  K @{term "(a :: int) + b"} ),
  ("prim_minus_int", K @{term "(a :: int) - b"} ),
  ("prim_mul_int",   K @{term "(a :: int) * b"} ),
  ("prim_div_int",   K @{term "(a :: int) div b"} ),
  ("prim_mod_int",   K @{term "(a :: int) mod b"} ),
  (* ("prim_and_int",   K @{term "(a :: int) AND b"} ), 
  ("prim_or_int",    K @{term "(a :: int) OR b"} ),
  ("prim_tilde_int", K @{term "NOT (a :: int)"} ), *)
  ("prim_incr_int",  K @{term "(a :: int) + 1"} ),
  ("prim_decr_int",  K @{term "(a :: int) - 1"} ),
  ("prim_eq_int",    K @{term "(a :: int) = b"} ),
  ("prim_neq_int",   K @{term "(a :: int) ~= b"} ),
  ("prim_less_int",  K @{term "(a :: int) < b"} ),
  ("prim_leq_int",   K @{term "(a :: int) <= b"} ),
  ("prim_greater_int", K @{term "b < (a :: int)"} ),
  ("prim_geq_int",     K @{term "b <= (a :: int)"} ),
  ("prim_plus_uint",   K @{term "(a :: nat) + b"} ),
  ("prim_minus_uint",  K @{term "(a :: nat) - b"} ),
  ("prim_mul_uint",    K @{term "(a :: nat) * b"} ),
  ("prim_div_uint",    K @{term "(a :: nat) div b"} ),
  ("prim_mod_uint",    K @{term "(a :: nat) mod b"} ),
(*  ("prim_and_uint",    K @{term "(a :: nat) AND b"} ),
  ("prim_or_uint",     K @{term "(a :: nat) OR b"} ),
  ("prim_tilde_uint",  K @{term "NOT (a :: nat)"} ),*)
  ("prim_incr_uint",   K @{term "(a :: nat) + 1"} ),
  ("prim_decr_uint",   K @{term "(a :: nat) - 1"} ),
  ("prim_eq_uint",     K @{term "(a :: nat) = b"} ),
  ("prim_neq_uint",    K @{term "(a :: nat) ~= b"} ),
  ("prim_less_uint",   K @{term "(a :: nat) < b"} ),
  ("prim_leq_uint",    K @{term "(a :: nat) <= b"} ),
  ("prim_greater_uint", K @{term "(a :: nat) > b"} ),
  ("prim_geq_uint",     K @{term "(a :: nat) >= b"} ),
 (*
  ("prim_plus_char",    K @{term "(a :: Word) + b"}  ),
  ("prim_minus_char",   K @{term "(a :: Word) - b"}  ),
  ("prim_mul_char",     K @{term "(a :: Word) * b"}  ),
  ("prim_div_char",     K @{term "(a :: Word) div b"}  ),
  ("prim_mod_char",     K @{term "(a :: Word) mod b"}  ),
  ("prim_incr_char",    K @{term "(a :: Word) + 1"} ),
  ("prim_decr_char",    K @{term "(a :: Word) - 1"} ),
  ("prim_tilde_char",   K @{term "NOT (a :: Word)"} ),
  ("prim_eq_char",      K @{term "(a :: Word) = b"} ),
  ("prim_neq_char",     K @{term "(a :: Word) ~= b"} ),
  ("prim_less_char",    K @{term "(a :: Word) < b"} ),
  ("prim_leq_char",     K @{term "(a :: Word) <= b"} ),
  ("prim_greater_char", K @{term "(a :: Word) > b"} ),
  ("prim_geq_char",     K @{term "(a :: Word) >= b"} ),
*)
 
  ("prim_not",          K @{term "Not (a :: bool)"} ),

(*                
  ("prim_plus_ptr", casePtrBinVoid @{term "(a :: Word) + sz * b"}),
*)
(*
  ("prim_plus_ptr_unsigned", K @{term False} ),
  ("prim_minus_ptr",         K @{term False} ),
  ("prim_incr_ptr",         K @{term False} ),
  ("prim_decr_ptr",         K @{term False} ),
*)
  ("prim_eq_ptr",           K @{term "(a :: addr) = b"} ),
  ("prim_neq_ptr",          K @{term "(a :: addr) ~= b"} )
 (* ("prim_less_ptr",         K @{term "(a :: Word) < b"} ),
  ("prim_leq_ptr",          K @{term "(a :: Word) <= b"} ),
  ("prim_greater_ptr",      K @{term "(a :: Word) > b"} ),
  ("prim_geq_ptr",          K @{term "(a :: Word) >= b"} ),
   *)             
 (*
 ("prim_tilde_char",
  { args = [ "TChar" ], res = "TChar",
    embed = fn [a] =>  to_Char (NOT (to_char a)) }),
 ("conv_int_char",
  { args = [ "TInt" ], res = "TChar",
    embed = fn [a] =>  to_Char (scast (to_int a)) }),
 ("conv_char_int",
  { args = [ "TChar" ], res = "TInt",
    embed = fn [a] =>  to_Int (scast (to_char a)) }),
 ("conv_uint_char",
  { args = [ "TUInt" ], res = "TChar",
    embed = fn [a] =>  to_Char (scast (to_uint a)) }),
 ("conv_char_uint",
  { args = [ "TChar" ], res = "TUInt",
    embed = fn [a] =>  to_UInt (scast (to_char a)) }),
*)
(*
 ("cast_int_ptr",  K @{term "(a :: Word)"} ),
 ("cast_ptr_int",  K @{term "(a :: Word)"} ),
 ("cast_int_uint", K @{term "(a :: Word)"} ),
 ("cast_uint_int", K @{term "(a :: Word)"} )
 *)
]

val primop_tab =
  let fun prep (nm, e) =
      (* TODO: pull out preparation of term to speed up *)
      (nm,
       fn (ctx,arg_tys) => fn args =>
         let val e = e (ctx,arg_tys)
            val frees =
                (* make sure that parameters are independent of the order in the term,
                   which may be surprising due to abbreviations *)
                sort (string_ord o pairself fst) 
                         (Term.add_frees e [])
          val closed = fold_rev absfree frees e
         in Term.list_comb (closed,args) end)
  in
      Symtab.make (map prep raw_tab)
  end

fun embed_primop ctx primop arg_tys args =
  case Symtab.lookup primop_tab primop of
      NONE => error ("embedding of primitive operator "^primop^" not done")
    | SOME f => Envir.beta_norm (f (ctx,arg_tys) args)

(**
 Types -- also probably
*)
open TypedAst
open TraceAux

fun embed_type_HOL ctx TInt = @{typ int}
  | embed_type_HOL ctx TUInt = @{typ nat}
  | embed_type_HOL ctx (TArray (t,i)) = Type(@{type_name array}, [ embed_type_HOL ctx t ])
  | embed_type_HOL ctx (TStructField t) = Type("fun", [ @{typ addr}, embed_type_HOL ctx t ])
  | embed_type_HOL ctx (TPtr t) = @{typ addr}
  | embed_type_HOL ctx (TStruct _) = error "cannot use structs as values in Burstall's model"
  | embed_type_HOL ctx t = error ("unexpected type "^PolyML.makestring t)
  
(**
  State space utils
*)
fun add_state_space name parents fun_name_opt decls extra thy =
  let val statetypeN = "StateType";
      val ctx = Proof_Context.init_global thy
      val vctx0 = vctx_mk fun_name_opt decls
      fun prep (x,(_,CVar ty)) =
          (var_name vctx0 x, embed_type_HOL ctx ty)
        | prep (x,(_,GhostVar ty)) =
          (var_name vctx0 x, ty)
      val comps = map prep decls @ extra
      val _ =
         trace_list ("in "^name^" declaring components")
            (map (fn (x,T) => Pretty.block [ Pretty.str x,
                                             Pretty.str " :: ",
                                             Syntax.pretty_typ ctx T ])
                comps)
  in
      StateSpace.define_statespace_i
          (SOME statetypeN) (* follow hoare_syntax / hoare? *)
          [] (* args *) name (* name *) 
          (map (fn p => (("",false),([],p,[]))) parents)
          comps thy (* the components *)
  end

val base_state_name = "base_state"
      
(** create global names *)
fun add_base_state_space thy =
  let val heap = Hoare.varname heap_var
      val thrown = Hoare.varname thrown_name 
      val heap = (if is_sep_logic_global thy
                  then [ (heap,@{typ heap}) ]
                  else [])
  in
      (thy |> add_state_space base_state_name [] NONE [] 
              (heap @ [(thrown, @{typ throw_case})])
          |> add_state_space (base_state_name^"_globals") [] NONE []
               (heap @ [(thrown, @{typ throw_case})]))
      handle ERROR _ => thy
  end 


fun open_locale name lthy =
  let val thy = Proof_Context.theory_of lthy
  in
      Locale.init (Locale.intern thy name) thy
  end

(**
  Accessing variables
*)
fun sK t = Abs ("_", dummyT, incr_boundvars 1 t);

fun is_state_comp ctx vctx name =
  case try (var_name vctx) name of
    NONE => false
    | SOME nm =>
      is_some (StateSpace.get_comp (Context.Proof ctx) nm)

val global_register =
   (cnst @{const_name StateSpace.globals},
    cnst @{const_name StateSpace.globals_update})
    
val local_register =
   (cnst @{const_name StateSpace.locals},
    cnst @{const_name StateSpace.locals_update})

fun var_register vctx v =
  if is_global_var vctx v
  then global_register
  else local_register

fun read_var ctx vctx v =
  let val (sel,_) = var_register vctx v
  in
      Abs("s",dummyT,
          StateSpace.gen_lookup_tr ctx (sel $ Bound 0) (var_name vctx v))
  end

fun set_var ctx vctx v t =
  let val (sel,upd) = var_register vctx v
  in
      Abs("s",dummyT,
          upd $ 
             (sK (StateSpace.gen_update_tr false ctx (var_name vctx v) 
                    (Term.incr_boundvars 1 t) (sel $ (Bound 0)))) $ Bound 0)
  end
  
fun replace_globals ctx src =
  let val (sel,upd) = global_register
  in Abs("s",dummyT,
         upd $ (sK (sel $ Term.incr_boundvars 1 src)) $ Bound 0)
  end  

(* Constructors for common terms *)
val sFalse = cnst @{const_name False}
val sTrue = cnst @{const_name True}
val sUndefined = cnst @{const_name undefined}
fun sCollect P = (Const(@{const_name Collect},dummyT)) $ P

val sSkip = cnst @{const_name Skip}
fun sBasic b = cnst @{const_name Basic} $ b
fun sSeq a b = cnst @{const_name Seq} $ a $ b
fun sLet a b = cnst @{const_name Let} $ a $ b
val sSkip = cnst @{const_name Skip}
fun sGuard f g a = cnst @{const_name Guard} 
                     $ f 
                     $ (sCollect g) $ a
fun sBind a b = cnst @{const_name bind} $ a $ b
fun sCond a b c = cnst @{const_name Cond} $ a $ b $ c
fun sWhile I e b = cnst @{const_name whileAnno} $ e $ I $ sUndefined $ b
fun sCatch a b = cnst @{const_name Catch} $ a $ b
val sThrow = cnst @{const_name Throw}
fun sBlock init body ret c = cnst @{const_name block} $ init $ body $ ret $ c
fun sCall init p ret c = cnst @{const_name call} $ init $ p $ ret $ c

fun s_throw ctx vctx which =
  sSeq (sBasic (set_var ctx vctx thrown_name which)) sThrow

fun s_catch ctx vctx which a b =
  sCatch a
    (sCond 
        (sCollect 
            (Abs("s",dummyT,
                 HOLogic.eq_const @{typ throw_case} 
                    $ (read_var ctx vctx thrown_name $ Bound 0)
                    $ which)))
        b
        sThrow)

fun mk_local (x,ty) = (x, (Local, CVar ty))
fun mk_global (x,ty) = (x, (Global, CVar ty))
fun mk_ghost_local (x,ty) = (x, (Local, GhostVar ty))
fun mk_ghost_global (x,ty) = (x, (Global, GhostVar ty))

fun mk_locals ds = map mk_local ds
fun mk_globals ds = map mk_global ds  
fun mk_ghost_locals ds = map mk_ghost_local ds
fun mk_ghost_globals ds = map mk_ghost_global ds
  
fun dest_name (Free(nm,_)) = nm
  | dest_name (Const(nm,_)) = Long_Name.base_name nm

fun dest_name_safe t =
  SOME (dest_name t)
  handle ERROR _ => NONE


fun fix_var_name "ret" = return_name
  | fix_var_name "heap" = heap_var
  | fix_var_name v = v

fun get_term_aux_vars ts =
  fold Term.add_frees ts [] |>
  filter (not o Hoare.is_state_var o fst) |>
  map fst

fun aCheck i a = cnst @{const_name acheck} $ i $ a
fun aGet i a = cnst @{const_name aget} $ a $ i
fun aSet i x a = cnst @{const_name aset} $ i $ x $ a

(**
  Struct fields in the Burstall model
*)
fun alloc_var s = s ^ "_alloc"
fun field_var s f = s ^ "_" ^ f

fun field_get ctx vctx s f p =
  Abs("s",dummyT, 
      (Const(@{const_name "field_get"},dummyT)
          $ (read_var ctx vctx (field_var s f) $ Bound 0)
          $ incr_boundvars 1 p))
  
fun field_put ctx vctx s f p v =
  Abs("s", dummyT, 
      set_var ctx vctx (field_var s f)
         (Const(@{const_name field_upd},dummyT) 
             $ (read_var ctx vctx (field_var s f) $ Bound 0)
             $ incr_boundvars 1 p
             $ incr_boundvars 1 v) $ Bound 0)
  
fun field_acc_cond ctx vctx s p =
  Abs("s",dummyT,
      (read_var ctx vctx (alloc_var s) $ Bound 0)
         $ incr_boundvars 1 p)

fun mk_field_var ctx s =
  let val { fields, ... } = get_struct_info (Proof_Context.theory_of ctx) s
  in
      mk_ghost_global (alloc_var s, Type("fun", [ @{typ addr}, @{typ bool}])) ::
      map (fn (f,ty) => mk_global (field_var s f, TStructField ty)) fields
  end

fun mk_field_vars ctx = maps (mk_field_var ctx)

val mk_alloc_ghosts = map (fn s => (alloc_var s, @{typ bool}))

(**
 Heap accesses for the separation logic heap model
*)
fun embed_type ctx TInt = @{const cty_int}
  | embed_type ctx (TPtr t) = @{const cty_ptr}
  | embed_type ctx TVoid = @{const cty_void}
  | embed_type ctx (TStruct nm) =
    (Consts.read_const (Proof_Context.consts_of ctx) ("cty_struct_"^nm, Position.none)
     handle ERROR _ => error ("embed_type: undefined struct type "^nm))
  | embed_type ctx ty = error ("Unexpected type: "^string_of_ty ty)

fun heap_put ctx vctx t p v =
   Abs("s",dummyT,
     set_var ctx vctx heap_var
       (Const(@{const_name heap_put},dummyT) 
           $ embed_type ctx t 
           $ incr_boundvars 1 p 
           $ incr_boundvars 1 v 
           $ (read_var ctx vctx heap_var $ Bound 0))
          $ Bound 0)
         

fun heap_get ctx vctx t p =
  Abs("s",dummyT,
   (Const(@{const_name heap_get},dummyT) $ embed_type ctx t 
         $ incr_boundvars 1 p
         $ (read_var ctx vctx heap_var $ Bound 0)))
  
fun heap_acc_cond ctx vctx t p =
  Abs("s",dummyT,
      (Const(@{const_name heap_acc_cond},dummyT) $ embed_type ctx t 
         $ incr_boundvars 1 p
         $ (read_var ctx vctx heap_var $ Bound 0)))

fun get_field_addr ctx p s f =
  let val intern = Consts.intern (Proof_Context.consts_of ctx)
  in
      (Const(@{const_name addr_add},dummyT) 
          $ p $ (Const(intern ("fo_"^s^"_"^f),dummyT)))
  end


(**
  Assertions are parsed strings and must therefore be read
  as logic terms first. The current state is accessed
  - if (free) variables occur
  - if the pattern a->f occurs
  - if the pattern a[i] occurs
  
  The ts is a (string * typ option) list
*)
fun embed_HOL_list ctx vctx state_depth ts =
  let (* HACK: we need to un-fix these for repeating type inference,
               because we modified the term structure after the
               initial type check was done *)
      val unfix_tvars =
      Term.map_types (Term.map_atyps (fn (TFree(x,S)) =>
                                         (TVar(("?"^x,0),HOLogic.typeS))
                                         | T => T))
      fun var d t =
        if dest_name t = ("\\"^"<sigma>")
        then Bound d
        else
        let val var_name = fix_var_name (dest_name t)
        in
            if is_state_comp ctx vctx var_name
            then read_var ctx vctx var_name $ Bound d
            else t
        end
      
      fun loop d (t as Const _) = var d t 
        | loop d (t as Free _) = var d t 
        (*| loop d (Const(@{syntax_const "@prestate"},_) $ (t as Const _)) = 
          var d t true
        | loop d (Const(@{syntax_const "@prestate"},_) $ (t as Free _)) =
          var d t true *)
        (*| loop d (Const(@{syntax_const "_faccess"},_) $ t $ f) =
           (case dest_name_safe t of
                NONE => (*TODO*) error("Sorry: struct access done only for pointer variables")
              | SOME v =>
                (case lookup_var ctx vctx vs v of
                    NONE => error ("unknown variable in field access: "^v)
                   | SOME(_, TPtr (ty as (TStruct s))) =>
                     let val f = dest_name f
                         val ft = get_struct_field_type ctx s f
                         val addr = field_addr ctx (read_var ctx vctx v $ Bound d) ty f
                     in 
                         heap_get ctx vctx ft addr $ Bound d
                     end
                   | SOME _ => error ("variable "^v^" is not pointer to struct"))) *)
        | loop d (a $ b) = loop d a $ loop d b
        | loop d (Abs(x,T,t')) = Abs(x,T, loop (d+1) t')
        | loop d t = t
      val ts' = 
         ts 
         |> (map (fn (t,T) =>
                    let val t' = Syntax.parse_term ctx t
                    in
                        case T of
                            NONE => t'
                          | SOME T' => Type.constraint T' t'
                    end))
         |> Syntax.check_terms ctx
         (* Syntax.read_terms ctx ts *)
      val aux = get_term_aux_vars ts'
  in  (ts' |>
       map (loop state_depth) |>
       map unfix_tvars, 
       aux)
  end
  
fun embed_HOL ctx vctx state_depth P = 
  apfst hd (embed_HOL_list ctx vctx state_depth [P])

fun embed_HOL_assn_list ctx vctx Ps =
  let val (Ps',aux) = embed_HOL_list ctx vctx 0 Ps
      val Ps'' = map (fn P' => sCollect (Abs("s", dummyT, P'))) Ps'
      (* val _ = trace_list "embedded" (map pretty_term_pure Ps'') *)
  in
      (Ps'',aux)
  end
  
fun embed_HOL_assn ctx vctx P =
  apfst hd (embed_HOL_assn_list ctx vctx [(P, SOME (@{typ bool}))])

(**
  The embedding of an expression yields a statement with a "hole", where
  the result of the expression is accessible as (Bound 0) and which
  must be filled by a statement. That statement is passed along as
  a continuation "k", both to the overall function and the loop.

  The VCG is slow if bind nodes are introduced too directly.
  We therefore use an intelligent bind function that understands
  the Simpl language and inspects the usages of the bound parameter.
  If they occur only in places where the state has not yet changed,
  the function inlines the references.  
*)

fun abs_k k = Abs("_",dummyT, k)

(*
 inc is the number of outer binders introduced besides this bind.
*)
fun bind ctx inc v k = 
  let exception StateChange
      open TraceAux
      (* b is the binding index, s the optional state index *)
      fun tr msg b t =
        tracing ("bind fail/"^msg^"/"^string_of_int b^
                    ":"^Syntax.string_of_term ctx t)
      
      fun check_occ msg b t =
        if member (op=) (loose_bnos t) b
        then (tr ("occ/"^msg) b t; raise StateChange)
        else ()
      
      fun repl b s (Bound i) =
         if i > b
         then Bound (i - 1)
         else if b = i 
         then (if s >= 0
               then betapply (incr_boundvars (s+1) v, Bound s)
               else (tr "repl" b (Bound i); raise StateChange))
         else Bound i
        | repl b s (Abs(x,T,t)) = Abs(x,T,repl (b+1) (s+1) t)
        | repl b s (u $ v) = repl b s u $ repl b s v
        | repl _ _ t = t
      
      fun inline b ((u as Abs _) $ v) =
         inline b (betapply (u,v))
        | inline b (Const(@{const_name Basic},_) $ f) =
          (Const(@{const_name Basic},dummyT) $ repl b ~1 f)
        | inline b (Const(@{const_name Seq},_) $ u $ v) =
          (check_occ "seq" b v;
           (Const(@{const_name Seq},dummyT) $ inline b u $ v))
        | inline b (Const(@{const_name Cond},_) $ e $ u $ v) =
          (Const(@{const_name Cond},dummyT) $ repl b ~1 e $ inline b u $ inline b v)
        | inline b (Const(@{const_name bind},_) $ e $ Abs(x,T,u)) =
          (Const(@{const_name bind},dummyT)
              $ repl b ~1 e 
              $ Abs(x,T, inline (b+1) u))
        | inline b (Const(@{const_name Guard},_) $ f $ g $ u) =
          (Const(@{const_name Guard},dummyT) $ f $ repl b ~1 g $ inline b u)
        | inline b (Const(@{const_name call},_) $ init $ p $ ret $ k) =
          (check_occ "call/p" b p;
           check_occ "call/ret" b ret;
           check_occ "call/k" b k;
           (Const(@{const_name call},dummyT) $ inline b init $ p $ ret $ k))
        | inline b t = (check_occ "other" b t; t)
  in
      inline 0 (incr_bv (inc,1,k))
      handle StateChange => sBind v (abs_k (incr_bv (inc+1,1,k)))
  end
  
(*
  Conservative check whether the short-circuit semantics 
  is redundant,  i.e. the rhs argument can be evaluated 
  regardless of the outcome of the lhs.
*)
val elim_short_circuit_redundant = Unsynchronized.ref true

fun short_circuit_redundant e =
    let fun loop (EVar _) = true
        | loop (ECchar _) = true
        | loop (ECint _) = true
        | loop (ECuint _) = true
        | loop ENull = true
        | loop (EDeref(_,_)) = false
        | loop (EAddr e) = loop e
        | loop (EArrow(e,_,_)) = false
        | loop (EAssign(_,a,b)) = false
        | loop (EIndex(a,i)) = false
        | loop (EPre(_,EVar _)) = true
        | loop (EPre(_,e)) = false
        | loop (EPost(_,EVar _)) = true
        | loop (EPost(_,e)) = false
        | loop (EPrim1(_,e)) = loop e
        | loop (EPrim2(_,a,b)) = loop a andalso loop b
        | loop (EFree(_,e)) = false
        | loop (EFreeArray(_,a,b)) = false
        | loop (ENew _) = true
        | loop (ENewArray(_,e)) = false
        | loop (EAnd(a,b)) = loop a andalso loop b
        | loop (EOr(a,b)) = loop a andalso loop b
        | loop (ECall(f,args)) = forall loop args
        | loop t = error ("not done: "^PolyML.makestring t)
  in
      !elim_short_circuit_redundant andalso loop e
  end

fun dest_TPtr (TPtr t) = t
  | dest_TPtr t = error ("dest_TPtr fails "^PolyML.makestring t)

fun dest_TStruct (TStruct s) = s
  | dest_TStruct t = error ("dest_TStruct fails "^PolyML.makestring t)

fun array_acc_check ctx vctx a i =
  Abs("s",dummyT, aCheck (incr_bv (1,0,i)) (read_var ctx vctx a $ Bound 0))                  

fun embed_exp ctx vctx e k =
  let val bind = bind ctx
      val bind0 = bind 0
      val bind1 = bind 1
      fun loop (EVar v) k = 
          bind0 (read_var ctx vctx v) k
        | loop (ECint i) k = 
          (* can use beta-reduction here *)
          betapply (abs_k k, param_HOL_number_of_int i)
        | loop ENull k =
          betapply (abs_k k, @{const NULL})
        | loop (EDeref (t,p)) k = 
          (case get_heap_model ctx of
               Burstall => 
               error "general dereferencing not allow in Burstall model"
             | SepLogic => 
               loop p (sGuard @{const HeapAlloc} 
                          (heap_acc_cond ctx vctx t (Bound 0))
                          (bind1 (heap_get ctx vctx t (Bound 0)) k)))
        | loop (EArrow (p, (TStruct s), f)) k =
          (case get_heap_model ctx of
              Burstall =>
               loop p
               (sGuard @{const HeapAlloc}
                   (field_acc_cond ctx vctx s (Bound 0))
                   (bind1 (field_get ctx vctx s f (Bound 0)) k))
             | SepLogic =>
               let val t = get_struct_field_type (Proof_Context.theory_of ctx) s f
                   val addr = get_field_addr ctx (Bound 0) s f
               in
                   loop p
                   (sGuard @{const HeapAlloc}
                       (heap_acc_cond ctx vctx t addr)
                       (bind1 (heap_get ctx vctx t addr) k))
               end)
        | loop (EIndex (EVar a, i)) k =
          loop i 
              (sGuard @{const ArrayBounds}
                   (array_acc_check ctx vctx a (Bound 0))                                          
                    (bind1 
                     (Abs("s",dummyT,
                          aGet (Bound 1) (read_var ctx vctx a $ Bound 0)))
                     k))
        | loop (EIndex _) k =
           error "arrays must be variables"
        | loop (EAssign(t,(EVar v), e)) k =
          loop e (sSeq (sBasic (set_var ctx vctx v (Bound 0))) 
                     (incr_bv (1,1,k)))
        | loop (EAssign (t, EArrow(p, TStruct s, f), e)) k =
          (case get_heap_model ctx of
            Burstall =>
               loop p
               (loop e
                   (sSeq (sGuard @{const HeapAlloc}
                             (field_acc_cond ctx vctx s (Bound 1))
                             (sBasic (field_put ctx vctx s f (Bound 1) (Bound 0))))
                       (incr_bv (2,0,k))))
           | SepLogic =>
               let val addr = get_field_addr ctx (Bound 1) s f
               in
                   loop p (loop e
                              (sSeq (sGuard @{const HeapAlloc}
                                        (heap_acc_cond ctx vctx t addr)
                                        (sBasic (heap_put ctx vctx t addr (Bound 0))))
                       (incr_bv (2,0,k))))
               end)
        | loop (EAssign(t, (EIndex (EVar a,i)), e)) k =
          (loop i
             (loop e
                (sGuard @{const ArrayBounds}
                   (array_acc_check ctx vctx a (Bound 1))
                   (sSeq (bind0 (read_var ctx vctx a)
                             (sBasic
                               (set_var ctx vctx a 
                                 (aSet (Bound 2) (Bound 1) (Bound 0)))))
                       (incr_bv (2,0,k))))))
        | loop (EAssign(t, (EDeref (t',p)), e)) k =
          (case get_heap_model ctx of
               Burstall =>
               error "general dereference not allowed in Burstall model"
             | SepLogic =>
               loop p (loop e 
                          (sSeq (sGuard @{const HeapAlloc}
                                    (heap_acc_cond ctx vctx t (Bound 1))
                                    (sBasic (heap_put ctx vctx t (Bound 1) (Bound 0))))
                              (incr_bv (2,0,k)))))
        | loop (t as (EAssign _)) k =
          error ("general assignment not supported: "^PolyML.makestring t)
        | loop (EPrim1((primop,ta),a)) k =
          (* use beta-reduction instead of bind for functional update *)
          loop a (betapply (abs_k (incr_bv (1,1,k)), embed_primop ctx primop [ta] [ Bound 0 ]))
        | loop (EPrim2((primop,ta,tb),a,b)) k =
          (* use beta-reduction instead of bind for functional update *)
          loop a (loop b (betapply (abs_k (incr_bv (2,1,k)), 
                                    embed_primop ctx primop [ta,tb] [ Bound 1, Bound 0])))
        | loop (EPost((primop,ta), EVar v)) k =
          bind0 (read_var ctx vctx v) 
               (sSeq (sBasic (set_var ctx vctx v 
                                 (embed_primop ctx primop [ta] [ Bound 0 ]))) k)
        | loop (EPre ((primop,ta), EVar v)) k =
          bind0 (read_var ctx vctx v) 
               (sSeq (sBasic (set_var ctx vctx v (embed_primop ctx primop [ta] [ Bound 0 ]))) 
                   (bind0 (read_var ctx vctx v) k))
        | loop (EAnd (a,b)) k =
          if short_circuit_redundant b
          then (loop a (loop b (betapply (abs_k (incr_bv (2,1,k)),
                                          (@{const HOL.conj} $ Bound 1 $ Bound 0)))))
          else (loop a (sCond (sCollect (sK (Bound 0)))
                           (loop b k)
                           (betapply (abs_k k, sFalse))))
        | loop (EOr (a,b)) k =
          if short_circuit_redundant b
          then (loop a (loop b (betapply (abs_k (incr_bv (2,1,k)),
                                          (@{const HOL.disj} $ Bound 1 $ Bound 0)))))
          else (loop a (sCond (sCollect (sK (Bound 0)))
                           (betapply (abs_k k, sTrue))
                           (loop b k)))
        | loop (ECall(f,args)) k =
          let val {params,rty,...} = get_fun_info ctx f
              val vctx' = vctx_mk (SOME f) (map mk_local 
                                               (params @ [(return_name,rty)]))
              val init =
                 Abs("s",dummyT,
                     fold (curry op $)
                        (map_index 
                            (fn (i,(p,_)) =>
                               set_var ctx vctx' p (Bound (i + 1)))
                            params)
                        (Bound 0))
              val ret =
                 Abs("s",dummyT,
                     Abs("t",dummyT,
                         set_var ctx vctx' return_name 
                            (read_var ctx vctx' return_name $ (Bound 0))
                            $ (replace_globals ctx (Bound 0) $ (Bound 1))))
          in
              fold loop args
              (sCall init (Free(suffix Hoare.proc_deco f,dummyT)) ret
                (Abs("_",dummyT, 
                     Abs("s",dummyT,
                         bind (length args+2)
                            (read_var ctx vctx' return_name) 
                               k))))
          end
        | loop e k = error("Embedding/exp not done: "^PolyML.makestring e)
  in
      loop e k
  end

fun contract (Const(@{const_name Seq},_) $ a $ Const(@{const_name Skip},_)) = 
    contract a
  | contract (Const(@{const_name Seq},_) $ Const(@{const_name Skip},_) $ b) =
    contract b
  | contract (Const(@{const_name Cond},_) 
                 $ (Const(@{const_name Collect},_) 
                       $ (Abs(_,_, (Const(@{const_name True},_)))))
                 $ u $ v) = contract u
  | contract (Const(@{const_name Cond},_) 
                 $ (Const(@{const_name Collect},_) 
                       $ (Abs(_,_, (Const(@{const_name False},_)))))
                 $ u $ v) = contract v
  | contract (a $ b) = contract a $ contract b
  | contract (Abs(x,T,t)) = Abs(x,T,contract t)
  | contract t = t
  
fun embed_sm ctx vctx syn =
  let fun loop vctx SSkip = sSkip
        | loop vctx (SSeq(a,b)) = 
          sSeq (loop vctx a) (loop vctx b)
        | loop vctx (SExec e) = 
          embed_exp ctx vctx e sSkip
        | loop vctx (SIf (e,a,b)) =
          embed_exp ctx vctx e
           (sCond (sCollect (sK (Bound 0))) (loop vctx a) (loop vctx b))
        | loop vctx (SWhile(SOME I,e,b)) =
          (s_catch ctx vctx @{const ThrowBreak}
              (sWhile (fst (embed_HOL_assn ctx vctx I)) (sCollect (sK sTrue))
                (embed_exp ctx vctx e
                  (sCond (sCollect (sK (Bound 0)))
                      (loop vctx b)
                      (s_throw ctx vctx @{const ThrowBreak}))))
              sSkip)
        | loop vctx (SWhile(NONE,_,_)) =
          error "While must have an invariant"
        | loop vctx (SBreak) =
          s_throw ctx vctx @{const ThrowBreak}
        | loop vctx (SBlock(decls,body)) =
          let val vctx = vctx_add_cur (map mk_local decls) vctx
              fun vars f = fold (fn d => fn r => f d $ r) decls (Bound 0)
              fun havoc_var (x,t) =
                 set_var ctx vctx x sUndefined
              val init = 
                 Abs("s",dummyT, vars havoc_var)
              fun rest_var (x,t) =
                set_var ctx vctx x (read_var ctx vctx x $ (Bound 1))
              val restore =
                 Abs("s",dummyT, Abs("t",dummyT, vars rest_var))                                      
          in
              sBlock init (loop vctx body) restore 
                 (sK (sK sSkip))
          end
        | loop vctx (SGhost(GAssign(l,t))) =
          let val l' = SimplC.parseExpression l
          in
              case l' of
                 Ast.EVar x =>
                  case vctx_get_cur vctx x of 
                      (_, GhostVar T) =>
                      sBasic (Abs("s",dummyT, 
                                  set_var ctx vctx x (fst (embed_HOL ctx vctx 0 (t, SOME T)))
                                     $ Bound 0))
                    | _ => error ("Not a ghost variable: "^x)
                | _ => error ("Sorry, complex ghost assignment not done: "^l)
          end
        | loop (vctx as VarContext {fun_name,...}) (SReturn e) =
          (case fun_name of
               NONE => error "Cannot embed return with function context"
             | SOME name =>
               embed_exp ctx vctx e 
                  (sSeq (sBasic (set_var ctx vctx return_name (Bound 0)))
                      (s_throw ctx vctx @{const ThrowReturn})))
        | loop vctx sm =
          error ("Embedding/sm not done: "^PolyML.makestring sm)
  in
      contract (loop vctx syn)
  end
 
 
fun list_all (xs, t) = 
  fold_rev (fn (f as Free(_,T)) => 
            fn P => HOLogic.all_const T $ lambda f P) xs t;

(**
  Type_Infer.prepare introduces new type parameters for
  constants. Make sure they do not clash with previously
  chosen type paramaters for auxiliary variables
*)
fun dummify_aux (@{const Trueprop} $ P) = @{const Trueprop} $ dummify_aux P
  | dummify_aux (Const(@{const_name HOL.All},_) $ Abs(x,_,t)) =
  (Const(@{const_name HOL.All},dummyT) $ Abs(x,dummyT,dummify_aux t))
  | dummify_aux t = t

fun triple ctx aux F pre sm post =
  let fun prep_type T =
        (tracing (PolyML.makestring T);
         T)
      val aux' = fold Term.add_frees [F,pre,sm,post] [] |>
                 filter (member (op=) aux o fst) |>
                 map Free
  in
 (@{term Trueprop} 
     $ list_all 
       (aux',
        Const(@{const_name "HoarePartialDef.hoarep"},dummyT)                   
           $ Free ("\<Gamma>", dummyT)
           $ Const(@{const_name bot},dummyT)
           $ F
           $ pre
           $ sm
           $ post
           $ Const(@{const_name bot},dummyT)))
  (*|> kill_auth*)
  (* |> dummify *)
  |> dummify_aux
  |> Envir.beta_eta_contract
  |> trace_filter (tracing o PolyML.makestring)
  |> trace_filter (trace_single "triple raw" o pretty_term_pure)
  |> singleton (snd o Type_Infer_Context.prepare ctx)
  |> trace_filter (trace_single "prepared" o pretty_term_pure)
  |> singleton (Type_Infer_Context.infer_types ctx #>
                Type_Infer.fixate ctx)
  |> trace_filter (trace_single "checked" o pretty_term_pure)
  end

(**
  Determine the accessed struct types
*)
fun used_structs_decls decls r =
  let fun in_ty TVoid r = r
        | in_ty TInt r = r
        | in_ty TUInt r = r
        | in_ty TChar r = r
        | in_ty (TPtr ty) r = in_ty ty r
        | in_ty (TArray(ty,_)) r = in_ty ty r
        | in_ty (TStruct s) r = insert (op=) s r
      
      fun in_decl (_, ty) r = in_ty ty r
  in
    fold in_decl decls r
  end
      
fun used_structs_sm sm r =
  let 
      fun loop SSkip r = r
        | loop (SExec _) r = r
        | loop (SSeq(a,b)) r = loop a (loop b r)        
        | loop (SBlock(decls,body)) r = loop body (used_structs_decls decls r)
        | loop (SIf(e,a,b)) r = loop a (loop b r)
        | loop (SWhile(I,e,a)) r = loop a r
        | loop SBreak r = r
        | loop SContinue r = r
        | loop (SReturn _) r = r
        | loop (SGhost _) r = r
  in
      loop sm r
  end

fun shadow_field_consts structs lthy =
  let fun shadow name lthy = 
        let val {name=_,fields} = 
               get_struct_info (Proof_Context.theory_of lthy) name
            fun field_binding (f,ty) = (Binding.name f, SOME (embed_type_HOL lthy ty), NoSyn)
        in
            snd (Proof_Context.add_fixes (map field_binding fields) lthy)
        end
  in 
      fold shadow structs lthy
  end

(**
  Top-level goals
*)
fun verify_statement (((((name,vars),ghost),pre),post),sm) lthy0 =
  let val var_decls = case vars of 
                          NONE => []
                        | SOME vars => SimplC.parseDecls vars
      val vars' = (mk_locals var_decls @
                   mk_ghost_locals (case ghost of
                                   SOME ghost => 
                                        (map (apsnd (Syntax.read_typ lthy0)) ghost)
                                  | NONE => []))
      val vctx = vctx_mk NONE vars'
      val ast = sm |>
                  SimplC.parseStatement |>
                  tc_sm (Proof_Context.theory_of lthy0) vctx
      val structs = used_structs_sm ast (used_structs_decls var_decls [])
      val _ = writeln ("Structs: "^commas structs)
      val vars'' = if is_burstall lthy0
                   then mk_field_vars lthy0 structs @ vars'
                   else vars'
      val _ = writeln (commas (map fst vars''))
      val vctx = vctx_mk NONE vars''
      
      val lthy0 = 
         Proof_Context.background_theory
                   (add_base_state_space #>
                    add_state_space name [base_state_name] NONE vars'' [])
                    lthy0
      
      val lthy = open_locale name lthy0
      val _ = writeln "Embedding statement ..."
      val sm' = ast |> embed_sm lthy vctx
      val _ = writeln "Embedding pre/post ..."
      
      val ([pre',post'], aux) = embed_HOL_assn_list lthy vctx [(pre, SOME @{typ bool}),
                                                                (post, SOME @{typ bool}) ]
      val goal =
          triple lthy aux (Const(@{const_name bot},dummyT)) pre' sm' post'

      fun after_qed [[verified]] lthy' =
        let val verified' = singleton (Proof_Context.export lthy' lthy0) verified
            val _ = trace_single "verified" (pretty_thm lthy verified')
        in lthy0 |>
           Local_Theory.note ((Binding.qualified_name name,[]),[verified']) |>
           snd
        end
  in
      Proof.theorem NONE after_qed [ [ (goal,[]) ] ] lthy
  end
  
val parse_ghost_vars =
  Scan.repeat1
    (Parse.name -- (Parse.$$$ "::" |-- Parse.typ))

val setup =
  heap_model_setup

val _ =
  Outer_Syntax.command
      @{command_spec "verify_statement"}
      "setup goal to verify a statement" 
      ((Parse.short_ident --
       (Scan.option (Args.$$$ "vars" |-- Parse.$$$ ":" |-- Parse.text)) --
       (Scan.option (Args.$$$ "ghost" |-- Parse.$$$ ":" |-- parse_ghost_vars)) --
       (Args.$$$ "pre" |-- Parse.$$$ ":" |-- Parse.term) -- 
       (Args.$$$ "post" |-- Parse.$$$ ":" |-- Parse.term) --
       Parse.text) >> 
        (Toplevel.local_theory_to_proof NONE o verify_statement))
end

