structure StructDef =
struct

open TraceAux
open TypedAst

fun cnst nm = Const(nm, dummyT)
fun lcnst ctx c = 
  Proof_Context.read_const_proper ctx false c
  |> dest_Const |> apsnd (K dummyT) |> Const

val embed_type_HOL = SimplC_Goals.embed_type_HOL
val embed_type = SimplC_Goals.embed_type

fun compose_fields [] f u prj = u
  | compose_fields fields f u prj =
     (Library.foldr1 (fn (t,r) => (f $ t $ r))
         (map_index (fn (i, (f,r)) => prj (i, (f,r))) fields))

fun cty_struct name = "cs_"^name
fun hol_rep_name sname = sname
fun field_const_name sname (f,_) = sname^"_"^f
fun field_offset_name sname (f,_) = "fo_"^sname^"_"^f
fun field_offset_const sname f = Const(field_offset_name sname f, @{typ nat})

fun HOL_representation {name,fields} thy =
  let val rec_fields =
         map (fn f => (Binding.name (field_const_name name f),
                       embed_type_HOL (Proof_Context.init_global thy) (snd f), NoSyn))
             fields
  in      
      Record.add_record ([],Binding.qualified_name (hol_rep_name name)) NONE rec_fields thy
  end
  
fun eval_number ctx t =
  let val eq = Simplifier.rewrite (simpset_of ctx)
                (cterm_of (Proof_Context.theory_of ctx) t)
      val _ = trace_list "evaluate" [ pretty_thm ctx eq ]
  in
     HOLogic.dest_nat (term_of (Thm.rhs_of eq))
     handle ERROR _ => error ("Evaluation failed: "^Display.string_of_thm ctx eq)
  end

fun simple_def ctx c_name rhs thy =
    let val rhs = singleton (Type_Infer_Context.infer_types ctx) rhs
        val ctx = Proof_Context.init_global thy
        val def_name = c_name^"_def"
	(* val _ = writeln ("Defining constant "^c_name) *)
        val ty = fastype_of rhs
        val (cnst, thy) = Sign.declare_const_global ((Binding.name c_name,ty),NoSyn) thy
        val eq = Logic.mk_equals (cnst,rhs)
	val ([d],thy) =
           Global_Theory.add_defs false
                          [((Binding.name def_name, eq), [])] thy
    in
        (cnst,d,thy)
    end
    
fun simple_note name thm thy =
  Global_Theory.add_thm ((Binding.name name,thm),[]) thy

val infer = singleton o Type_Infer_Context.infer_types

val old_infer = infer
fun infer ctx t =
  let val _ = write_list "infer" [ Syntax.pretty_term ctx t ]
  in old_infer ctx t
  end

fun define_offsets {name,fields} thy =
  let val ctx = Proof_Context.init_global thy
      fun add f (off,thy) =
        let val rhs = HOLogic.mk_number @{typ nat} off
            val (c,d,thy) = simple_def ctx (field_offset_name name f) rhs thy
            val _ = writeln (Display.string_of_thm ctx d)
            val delta = eval_number ctx
                            (infer ctx (cnst @{const_name ty_size} $ embed_type ctx (snd f)))
        in
            ((c,d),(off+delta,thy))
        end
      val (defs,(_,thy)) = fold_map add fields (0,thy)
  in 
      (defs,thy)
  end

fun add_sep_logic_support info thy =
  let val {name,fields} = info
            
      (* records use theories, so the following cannot be localized (yet) *)
      val thy = HOL_representation info thy
      val ctx0 = Proof_Context.init_global thy
      
      val as_ctx = Proof_Context.init_global
      
      val compose_fields = compose_fields fields
      
      val ctx = as_ctx thy
      val size =
         infer ctx
           (compose_fields (cnst @{const_name "plus"}) @{term "0 :: nat"} 
               (fn (_,(f,r)) =>  (cnst @{const_name ty_size}) $ embed_type ctx0 r))
        
      val _ = write_list "size is" [ pretty_term ctx0 size ]

      val frep = 
        infer ctx
          (Abs("r",dummyT,
               compose_fields (cnst @{const_name "append"}) (cnst @{const_name "Nil"})
                  (fn (_,f) => 
                     cnst @{const_name"ty_rep"} $ embed_type ctx0 (snd f) $
                      (lcnst ctx0 (field_const_name name f) $ Bound 0))))
      
      val _ = write_list "repr is" [ pretty_term ctx0 frep ]

      val (offset_defs,thy) = define_offsets info thy
      val ctx = as_ctx thy
      val _ = trace_list "offset_defs" (map (pretty_thm ctx0 o snd) offset_defs)
      val fval =
        infer ctx
          (Abs("v",dummyT,
              list_comb (lcnst ctx (hol_rep_name name ^ ".make"), 
                         map (fn (f,(c,_)) => 
                                let val ty = embed_type ctx (snd f)
                                in
                                    cnst @{const_name ty_val} $ ty
                                  $ (cnst @{const_name take} $ (cnst @{const_name ty_size} $ ty)
                                        $ (cnst @{const_name drop}
                                              $ c
                                              $ Bound 0))
                                 end)
                              (fields ~~ offset_defs))))
     val _ = write_list "val is" [ pretty_term ctx fval ]
    
     val (struct_cnst,def,thy) =
        simple_def ctx (cty_struct name) 
                   (cnst @{const_name "cty.make"} $ size $ frep $ fval) thy
        
      val _ = write_list "have type definition" [ pretty_thm ctx def ]
      val ctx = as_ctx thy

      val ty_ok = 
         let val simps = 
                Global_Theory.get_thm thy (name^".make_def") ::
                Global_Theory.get_thms thy (name^".simps")
             val goal = infer ctx (@{term Trueprop} $ (cnst @{const_name ty_ok} $ (lcnst ctx (cty_struct name))))
         in            
             Goal.prove ctx [] [] goal
             (fn _ => 
                simp_tac (HOL_basic_ss addsimps @{thms ty_ok_def}) 1 THEN
                simp_tac (simpset_of ctx addsimps
                             (simps @
                              [def] @
                              @{thms cty.make_def} @
                              @{thms ty_val_ty_rep_inv} @
                              @{thms tyrep_size_by_ty_ok} @ 
                               map snd offset_defs)) 1)
         end 
            
      val _ = trace_list "proved ty ok" [ pretty_thm ctx ty_ok ]  
      
      val (_,thy) = simple_note (cty_struct name^"_ty_ok") ty_ok thy
         
      val ty_size = Simplifier.rewrite (simpset_of ctx addsimps 
                                          [ @{thm cty.make_def}, def ])
                         (cterm_of thy (infer ctx (cnst @{const_name ty_size} $ struct_cnst)))
        
      val (_,thy) = simple_note (cty_struct name^"_ty_size") ty_size thy
        
      val _ = write_list "have ty_size" [ pretty_thm ctx ty_size ]
      
      val thy =
        Context.theory_map 
             (Simplifier.map_ss (fn ss => ss addsimps [ ty_ok, ty_size ]))
             thy
      
      val ctx = as_ctx thy
      
      (* Prove the unfolding theorem for separation logic reasoning *)
      
      val n_fields = length fields
      fun bind_Ex x t = cnst @{const_name heap_ex} $ Abs(x,dummyT,t)

      fun spatial p v =
         fold bind_Ex (rev (map fst fields))
          (@{const heap_conj}
              $ (compose_fields @{const star} @{const emp}
                    (fn (i, f) => 
                       (cnst @{const_name points_to} 
                           $ (cnst @{const_name addr_add} $ p $ lcnst ctx (field_offset_name name f))
                           $ embed_type ctx (snd f)
                           $ Bound (n_fields - i - 1))))
              $ (cnst @{const_name HOL.eq} $ v $ 
                  (list_comb (lcnst ctx (hol_rep_name name ^ ".make"),
                              rev (map_range (fn i => Bound i) n_fields)))))
      
      val make_def =
         Global_Theory.get_thm thy (hol_rep_name name ^ ".make_def")
     
      val unfold_make =
         Pattern.rewrite_term thy [Logic.dest_equals (prop_of make_def)] []
      val tmp = (Abs("p", dummyT, Abs ("v", dummyT, spatial (Bound (n_fields +1)) (Bound n_fields))))
      val _ = writeln (Syntax.string_of_term ctx tmp)
      val _ = writeln "B"
      val rhs = 
         infer ctx tmp 
        |> unfold_make

      val _ = write_list "seplogic rhs" [ pretty_term ctx rhs ]

      val (scnst,sdef,thy) = simple_def ctx name rhs thy
      val ctx = as_ctx thy
      val _ = write_list "seplogic def is" [ pretty_thm ctx sdef ]
      
      val unfold_eq =
        let val p = Free("p",dummyT)
            val v = Free("v",dummyT)
            val g = 
               (@{term Trueprop} $ (cnst @{const_name HOL.eq} $ (scnst $  p $ v) $ spatial p v))
               |> infer ctx
               |> unfold_make
        in
            Goal.prove ctx ["p","v"] [] g 
            (fn _ => simp_tac (HOL_basic_ss addsimps 
                                  [sdef,make_def]) 1)
        end
      val _ = write_list "unfold theorem" [ pretty_thm ctx unfold_eq ]
      
      val (_,thy)= simple_note (name^"_unfold") unfold_eq thy
      
      val thy = Context.theory_map (Heaps.add_unfold_eq false unfold_eq) thy
  in
      thy
  end

fun define_struct src thy =
  let val info = SimplC.parseStructDef src
      val _ = writeln ("defined struct " ^ #name info^" with "^
                          commas (map fst (#fields info)))
      val _ = TypedAst.check_struct_info thy info
  in
      thy |>
      TypedAst.put_struct_info info |>
      (if SimplC_Goals.is_sep_logic_global thy
       then add_sep_logic_support info
       else I)
  end

val _ =
  Outer_Syntax.command
      @{command_spec "define_struct"}
      "define a new struct type" 
      (Parse.text >> (Toplevel.theory o define_struct))

end