structure TraceAux =
struct

(* Pretty printing *)
fun pretty_cterm ctx = Syntax.pretty_term ctx o Thm.term_of;
fun pretty_ctyp ctx = Syntax.pretty_typ ctx o #T o Thm.rep_ctyp
fun pretty_term ctx = Syntax.pretty_term ctx
fun pretty_typ ctx = Syntax.pretty_typ ctx

fun pretty_thm ctx =
  Display.pretty_thm_raw ctx {quote=false,show_hyps=false}
fun pretty_param ctx (nm,T) =
  Pretty.block [ Pretty.str nm, Pretty.str ":", pretty_typ ctx T ]

val pretty_term_pure =
  pretty_term (Proof_Context.init_global Pure.thy)

fun pretty_params ctx ps =
  Pretty.separate " " (map (pretty_param ctx) ps)

val string_of_term = Pretty.string_of oo pretty_term
val string_of_cterm = Pretty.string_of oo pretty_cterm
val string_of_thm = Pretty.string_of oo pretty_thm
val string_of_params = Pretty.string_of o Pretty.block oo pretty_params

fun trace_list msg ps =
  tracing (Pretty.string_of (Pretty.big_list msg ps))

fun trace_thm ctx msg thm =
  (trace_list msg [ pretty_thm ctx thm ]; thm)

fun trace_term ctx msg t =
  (trace_list msg [ pretty_term ctx t ]; t)

fun trace_conv ctx msg ct =
  (tracing (Pretty.string_of
               (Pretty.block [ Pretty.str (msg^": "),
                               pretty_cterm ctx ct ]));
   Conv.all_conv ct)

fun trace_prem_tac ctx msg i st =
  (trace_list msg [ pretty_cterm ctx (cprem_of st i) ];
   Seq.single st)

val string_of_list = Pretty.string_of oo Pretty.big_list

fun trace_filter f t = (f t; t)

fun trace_single msg p = 
  trace_list msg [ p ]

fun write_list msg ps =
  writeln (Pretty.string_of (Pretty.big_list msg ps))

end